/* IR.cpp
    Program for calculating Infra red spectrum from the MD coordinates
---------------------------------------------------------------------------*/
/* USAGE
    $ g++ IR.cpp
    $ ./a.out
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
    Copyright (c) 2023 The Society of Polymer Science, Japan

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
---------------------------------------------------------------------------*/

#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>
#include <cmath>
#include <complex>

#define TOTAL_TIME 52.4288E+00// [psec]
#define PSEC2SEC 1.0E-12
#define SPEED_OF_LIGHT 2.99792458E+8 // [m/sec]

void FastFourierTransform(int ndata, double *real_part, double *imaginary_part, int bforward){
    int power = 0;
    for(int i = 2; i <= ndata; i *= 2) power++;

    int inversion[ndata];
    inversion[0] = 0;
    int b = 1;
    for(int i = 0; i < power; i++){
        int m = ndata / (pow(2.0, i + 1));
        for(int j = b; j < b * 2; j++){
            inversion[j] = inversion[j - b] + m;
        }
        b *= 2;
    }
    for(int i = 0; i < ndata; i++){
        if(i < inversion[i]){
            double temp = real_part[i];
            real_part[i] = real_part[inversion[i]];
            real_part[inversion[i]] = temp;
            temp = imaginary_part[i];
            imaginary_part[i] = imaginary_part[inversion[i]];
            imaginary_part[inversion[i]] = temp;
        }
    }

    for(int i = 1; i < ndata; i *= 2){
        for(int j = 0; j < i; j++){
            double theta = -M_PI * j / i;
            double coeff[2];
            if(bforward == 0) theta *= -1;
            coeff[0] = cos(theta);
            coeff[1] = sin(theta);
            for(int k = 0; k < ndata; k += i * 2){
                double next[4];
                next[0]      = real_part[j + k];
                next[1]      = imaginary_part[j + k];
                next[2]      = real_part[j + k + i] * coeff[0] - imaginary_part[j + k + i] * coeff[1];
                next[3]      = real_part[j + k + i] * coeff[1] + imaginary_part[j + k + i] * coeff[0];
                real_part[j + k]          = next[0] + next[2];
                imaginary_part[j + k]     = next[1] + next[3];
                real_part[j + k + i]      = next[0] - next[2]; 
                imaginary_part[j + k + i] = next[1] - next[3];
            }
        }
    }
}

double *AutoCorrelation(int ndata, double *data){

    // Search N of 2**N and ZERO PADING
    int N = 2;
    while(N < ndata) N *= 2;
    N *= 2;
    double *fft_real, *fft_imaginary;
    fft_real = new double[N];
    fft_imaginary = new double[N];
    for(int i = 0; i < ndata; i++){
        fft_real[i] = data[i];
        fft_imaginary[i] = 0.0;
    }
    for(int i = ndata; i < N; i++){
        fft_real[i] = 0.0;
        fft_imaginary[i] = 0.0;
    }
    FastFourierTransform(N, fft_real, fft_imaginary, 1);

    for(int i = 0; i < N; i++){
        fft_real[i] = fft_real[i] * fft_real[i] + fft_imaginary[i] * fft_imaginary[i];
        fft_imaginary[i] = 0.0;
    }

    FastFourierTransform(N, fft_real, fft_imaginary, 0);

    double *Corr = (double*)malloc(sizeof(double) * ndata);
    for(int i = 0; i < ndata; i++){
        fft_real[i] /= N;
        Corr[i] = 0.0;
    }
    for(int i = 0; i < ndata; i++){
        Corr[i] = fft_real[i] / (ndata - i);
    }

    delete[] fft_real;
    delete[] fft_imaginary;

    return Corr;
};

double *IR(int ndata, int DIM, std::vector<std::vector<double> > data){

    double *auto_corr = (double *) malloc(sizeof(double) * ndata);
    // Zero clear
    for(int idata = 0; idata < ndata; idata++){
        auto_corr[idata] = 0.0;
    }

    // Calc. auto correlation function of derivated dipole of the system
    for(int ix = 0; ix < DIM; ix++){
        double *temp_auto_corr = (double *) malloc(sizeof(double) * ndata);
        double *data_1d = (double *) malloc(sizeof(double) * ndata);
        for (int idata = 0; idata < ndata; idata++){
            data_1d[idata] = data[idata][ix];
        }
        temp_auto_corr = AutoCorrelation(ndata, data_1d);
        for (int idata = 0; idata < ndata; idata++){
            auto_corr[idata] += temp_auto_corr[idata] / DIM;
        }
        free(temp_auto_corr);
        free(data_1d);
    }

    // Fourier transform of auto corr.
    int N = 2;
    while(N < ndata) N *= 2;

    double *fft_real, *fft_imaginary;
    fft_real = new double[N];
    fft_imaginary = new double[N];
    for(int i = 0; i < ndata; i++){
        fft_real[i] = auto_corr[i];
        fft_imaginary[i] = 0.0;
    }
    for(int i = ndata; i < N; i++){
        fft_real[i] = 0.0;
        fft_imaginary[i] = 0.0;
    }
    FastFourierTransform(N, fft_real, fft_imaginary, 1);

    double *spectrum = (double *) malloc(sizeof(double) * ndata);

    for(int idata = 0; idata < ndata; idata++){
        spectrum[idata] = sqrt(fft_real[idata] * fft_real[idata] + fft_imaginary[idata] * fft_imaginary[idata]);
    }

    delete[] fft_real;
    delete[] fft_imaginary;

    return spectrum;
}

int main(int argc, char** argv){
    std::string input_file_name = argv[1];
    std::vector<std::vector<double> > data;
    std::vector<double> time;
    int ndata;
    std::ifstream ifs(input_file_name);
    if (!ifs){
        std::cerr << "\a Can't open the file: " <<  input_file_name << std::endl;
    }
    else{
        // read data
        std::string line;
        ndata = 0;
        while (getline(ifs, line)){
            if(line.c_str()[0] == '#') continue;

            int i;
            double x, y, z;
            sscanf(line.c_str(), "%d%lE%lE%lE", &i, &x, &y, &z);
            std::vector<double> one_data;
            one_data.push_back(x);
            one_data.push_back(y);
            one_data.push_back(z);
            time.push_back((double) i);
            data.push_back(one_data);
            ndata++;
        }
    }

    // Calculate IR spectrum
    double *spectrum = (double *) malloc(sizeof(double) * ndata);
    spectrum = IR(ndata, 3, data);

    FILE *fp;
    fp = fopen("result.dat", "w");
    // Write result
    double delta_omega = 1.0 / (TOTAL_TIME * PSEC2SEC) / (SPEED_OF_LIGHT * 100.0); //[cm^-1]
    fprintf(fp, "# ");
    for(int i = 0; i < argc; i++){
        fprintf(fp, "%s ",argv[i]);
    }
    fprintf(fp, "\n");
    fprintf(fp, "# omega [cm^-1]  Spectrum [arb.unit]\n");
    for(int idata = 1; idata < ndata; idata++){
       double omega = idata * delta_omega;  //[cm^-1]
       fprintf(fp, "%15lf %20.10E\n", omega, spectrum[idata]);
    }
    fclose(fp);

    return 0;
}
